/*
TreeSplicer, a framework to enhance temporal locality in repeated tree traversals.
Youngjoon Jo (yjo@purdue.edu)
https://sites.google.com/site/treesplicer/

Copyright 2012, School of Electrical and Computer Engineering, Purdue University.

These benchmarks are adapted from the Lonestar benchmark suite:
http://iss.ices.utexas.edu/?p=projects/galois/lonestar
 */

package nearestneighbor;

public class Point {

	final static int DIM = KdCell.DIM;
	double [] coord;
	Point naiveClosest = null;
	Point closest = null;
	double closestDist = Double.MAX_VALUE;
	int label;
	
	int nodesTraversed = 0;

	public Point(int label, double [] coord) {
		this.label = label;
		this.coord = coord;
	}

	public double get(int i) {
		return coord[i];
	}

	public double [] getCoords() {
		return coord;
	}  

	public double dist(Point dest) {
		double sum = 0;
		double dist;
		for (int i = 0; i < DIM; i++) {
			dist = coord[i] - dest.coord[i];
			sum += dist * dist;
		}
		return sum;
	}

	public void updateClosest(Point dest) {
		if (dest == this) {
			return;
		}
		if (closest == null) {
			closest = dest;
			closestDist = dist(dest);
		} else {
			double newDist = dist(dest);	  
			if (newDist < closestDist) {
				closest = dest;
				closestDist = newDist;
			}
		}
	} 
}
