/*
TreeSplicer, a framework to enhance temporal locality in repeated tree traversals.
Youngjoon Jo (yjo@purdue.edu)

Copyright 2011, School of Electrical and Computer Engineering, Purdue University.

These benchmarks are adapted from the Lonestar benchmark suite:
http://iss.ices.utexas.edu/?p=projects/galois/lonestar
 */

package nearestneighborquad;

import java.util.Random;
import util.Launcher;

public class Main {

	static final int DIM = KdCell.DIM;
	static final boolean verify = false;
	static boolean sort;

	public static void main(String[] args) {
		if (args.length != 3) {
			System.err.println("arguments: sort numTrainPoints numTestPoints");
			System.err.println("ex) false 10000 10000");
			System.exit(-1);
		}
		sort = Boolean.parseBoolean(args[0]);
		int numTrainPoints = Integer.parseInt(args[1]);
		int numTestPoints = Integer.parseInt(args[2]);
		Random rand = new Random(0);
		Point[] trainData = genData(numTrainPoints, rand);
		Point[] testData = genData(numTestPoints, rand);
		process(trainData, testData);
	}
	
	static Point[] genData(int count, Random rand) {
		
		Point[] points = new Point[count];
		double[] coord = new double[DIM];
		for(int i = 0; i < count; i++) {
			for(int j = 0; j < DIM; j++) {
				coord[j] = rand.nextDouble();
			}
			points[i] = new Point(coord);
		}
		return points;
	}	

	static void process(final Point[] trainData, final Point[] testData) {
		if (verify) {
			for (Point test : testData) {
				for (Point train : trainData) {
					test.updateClosest(train);
				}
				test.naiveClosest = test.closest;
				test.closest = null;
				test.closestDist = Double.MAX_VALUE;
			}
		}

		KdCell root = KdCell.createTree(trainData); 
		if (sort) {
			KdCell.createTree(testData);
		}

		Launcher.startTiming();
		ParallelFor(root, testData);
		Launcher.stopTiming();
		
		if (testData.length == 10000 && trainData.length == 10000) {
			// check expected values
			long totalNodesTraversed = 0;
			for (int i = 0; i < testData.length; i++) {
				totalNodesTraversed += testData[i].nodesTraversed;
			}
			System.out.printf("totalNodesTraversed: %d\n", totalNodesTraversed);
			assert totalNodesTraversed == 7298188L;
		}

		if (verify) {
			for (Point test : testData) {
				assert test.naiveClosest == test.closest;
			}
			System.out.println("verify successful");
		}		
	}
	
	// a substitute for parallel for annotations on for statements, which are not supported in Java 6
	public @interface ParallelFor {}
	@ParallelFor
	public static void ParallelFor(KdCell root, Point [] testData) {
		for (int i = 0 ; i < testData.length; i++) {
			Point test = testData[i];
			root.findNearestNeighbor(test);
		}
	}
}
